/*
 * SPDX-FileCopyrightText: 2026 Oak Ridge National Laboratory and Contributors
 *
 * SPDX-License-Identifier: Apache-2.0
 */

#ifndef ADIOS2_BINDINGS_C_ADIOS2_C_H_
#define ADIOS2_BINDINGS_C_ADIOS2_C_H_

#include "adios2/common/ADIOSConfig.h"

#include "adios2/c/adios2_c_adios.h"
#include "adios2/c/adios2_c_attribute.h"
#include "adios2/c/adios2_c_engine.h"
#include "adios2/c/adios2_c_io.h"
#include "adios2/c/adios2_c_operator.h"
#include "adios2/c/adios2_c_variable.h"

#endif /* ADIOS2_BINDINGS_C_ADIOS2_C_H_ */
