/*
 * SPDX-FileCopyrightText: 2026 Oak Ridge National Laboratory and Contributors
 *
 * SPDX-License-Identifier: Apache-2.0
 */

#ifndef ADIOS2_BINDINGS_CXX_CXX_ADIOS_VIEW_H_
#define ADIOS2_BINDINGS_CXX_CXX_ADIOS_VIEW_H_

namespace adios2
{
template <typename T, class... Parameters>
class AdiosView
{
public:
    AdiosView() = delete;
};
}

#endif /* ADIOS2_BINDINGS_CXX_CXX_ADIOS_VIEW_H_ */
